import os
from os.path import join, basename, splitext
import logging
from glob import glob
import shutil
from typing import List, Tuple

from ..util import gdal, osr, replace_ext


logger = logging.getLogger(__name__)


def georeference_step(source_dir: os.PathLike, target_dir: os.PathLike, preprocessor_config: dict, geotransforms: List[dict]):
    success = False
    for options in geotransforms:
        type_name = options['type'].lower()
        opts_dict = {i: options[i] for i in options if i != 'type'}
        if type_name == 'gcp':
            georef_func = gcp_georef
        elif type_name == 'rpc':
            georef_func = rpc_georef
        elif type_name == 'world':
            georef_func = world_georef
        elif type_name == 'corners':
            georef_func = corner_georef
        elif type_name == 'no_op':
            georef_func = no_op
        else:
            raise Exception('Invalid georeference type %s' % type_name)
        try:
            filenames = []
            for dataglob in preprocessor_config.get('data_file_globs', '*'):
                for p in [path for path in glob(join(source_dir, '**', dataglob), recursive=True) if not os.path.isdir(path)]:
                    filenames.append(p)
            for filename in filenames:
                target_filename = join(target_dir, basename(filename))
                georef_func(filename, target_filename, **opts_dict)
        except Exception as e:
            # if any file fails, try another method
            logger.warn(e)
            continue
        else:
            # all files georeferenced without exception, no need to try another methods
            success = True
            logger.debug('Used geotransform method success: %s' % type_name)
            break
    if not success and len(geotransforms) > 0:
        raise Exception('No configured georeference method successfull: %s' % ', '.join([item['type'] for item in geotransforms]))


def gcp_georef(input_filename: os.PathLike, target_filename: os.PathLike, order: int=1, projection: str='EPSG:4326',
               tps: bool=False, warp_options: dict=None):
    succeded = False

    # simple case: get the geotransform from some GCPs
    try:
        ds = gdal.Open(input_filename, gdal.GA_Update)
    except RuntimeError:
        logger.warn('Can not open file by GDAL %s' % (input_filename))
        return
    if ds.GetGCPCount() <= 4:
        try:
            gcps = ds.GetGCPs()
            gt = gdal.GCPsToGeoTransform(gcps)
            ds.SetGeoTransform(gt)
        except Exception:
            del ds
            logger.warning("Failed to get geotransform via GCPsToGeoTransform")
        else:
            del ds
            shutil.move(input_filename, target_filename)
            succeded = True
            logger.debug("GCPsToGeoTransform successful")

    # otherwise warp
    if not succeded:
        logger.info("Applying GCP transform by warping")

        if tps:
            options = {
                'tps': tps
            }
        else:
            options = {
                'polynomialOrder': order
            }

        gdal.Warp(
            target_filename,
            input_filename,
            dstSRS=projection,
            **(warp_options or {}),
            **options,
        )


def rpc_georef(input_filename: os.PathLike, target_filename: os.PathLike, rpc_file_template: str='{fileroot}.RPC', warp_options: dict=None):
    fileroot, extension = splitext(input_filename)
    rpc_file_glob = rpc_file_template.format(
        filename=input_filename, fileroot=fileroot,
        extension=extension,
    )
    rpc_filename = None
    try:
        rpc_filename = glob(rpc_file_glob, recursive=True)[0]
    except IndexError:
        logger.warn('No RPC filename found with glob %s' % rpc_file_glob)
    # rename RPC filename to be compatible with GDAL
    if rpc_filename:
        shutil.move(rpc_filename, replace_ext(input_filename, '.rpc'))
    gdal.Warp(
        target_filename,
        input_filename,
        rpc=True,
        **(warp_options or {})
    )


def corner_georef(input_filename: os.PathLike, target_filename: os.PathLike, corner_names: List[str]=None,
                  orbit_direction_name: str=None, force_north_up: bool=False, gcp_srid: int=4326, warp: bool=False):
    corner_names = corner_names or ["bottom_left", "bottom_right", "top_left", "top_right"]
    ds = gdal.Open(input_filename, gdal.GA_Update)

    orbit_direction = ds.GetMetadata()[orbit_direction_name].lower()
    metadata = ds.GetRasterBand(1).GetMetadata()

    # from pprint import pprint

    # pprint (metadata)
    # pprint(ds.GetMetadata())
    bl, br, tl, tr = [
        [float(num) for num in metadata[corner_name].split()]
        for corner_name in corner_names
    ]

    gcps = gcps_from_borders(
        (ds.RasterXSize, ds.RasterYSize),
        (bl, br, tl, tr),
        orbit_direction,
        force_north_up
    )

    sr = osr.SpatialReference()
    sr.ImportFromEPSG(gcp_srid)

    ds.SetGCPs(gcps, sr.ExportToWkt())

    if warp:
        gdal.Warp(
            target_filename,
            ds,
        )
        del ds
    else:
        ds.SetGeoTransform(gdal.GCPsToGeoTransform(ds.GetGCPs()))
        driver = ds.GetDriver()
        del ds
        driver.Rename(target_filename, input_filename)


def world_georef():
    # TODO: implement
    pass


def no_op(input_filename: os.PathLike, target_filename: os.PathLike):
    # does not perform any operation
    # configure as last step if you want other geotransform method to either pass or fail and then perform no other
    shutil.move(input_filename, target_filename)


def gcps_from_borders(size: Tuple[float, float], coords: List[Tuple[float, float]], orbit_direction: str, force_north_up: bool=False):
    x_size, y_size = size
    # expects coordinates in dict(.*border_left.*:[lat,lon],...)
    gcps = []
    if force_north_up and len(coords) == 4:
        # compute gcps assuming north-up, east-right image no matter, what is claimed by metadata
        sorted_by_lats = sorted(coords, key=lambda x: x[0], reverse=True)
        # compare longitudes
        if sorted_by_lats[0][1] > sorted_by_lats[1][1]:
            #                                                            /\
            #  1                                                         \ \
            # top - left corner has lower latitude from two northernmost  \/
            if orbit_direction != "descending":
                top_left = sorted_by_lats[1]
                top_right = sorted_by_lats[0]
                bottom_left = sorted_by_lats[3]
                bottom_right = sorted_by_lats[2]
            else:
                top_left = sorted_by_lats[3]
                top_right = sorted_by_lats[2]
                bottom_left = sorted_by_lats[1]
                bottom_right = sorted_by_lats[0]
        else:
            #                                                                /\
            #   2                                                           / /
            #  top - left corner has higher latitude from two northernmost  \/
            if orbit_direction != "descending":
                top_left = sorted_by_lats[0]
                top_right = sorted_by_lats[1]
                bottom_left = sorted_by_lats[2]
                bottom_right = sorted_by_lats[3]
            else:
                top_left = sorted_by_lats[2]
                top_right = sorted_by_lats[3]
                bottom_left = sorted_by_lats[0]
                bottom_right = sorted_by_lats[1]
        gcps.append(gdal.GCP(bottom_left[1], bottom_left[0], 0, 0.5, 0.5))
        gcps.append(gdal.GCP(bottom_right[1], bottom_right[0], 0, x_size - 0.5, 0.5))
        gcps.append(gdal.GCP(top_left[1], top_left[0], 0, 0.5, y_size - 0.5))
        gcps.append(gdal.GCP(top_right[1], top_right[0], 0, x_size - 0.5, y_size - 0.5))

    else:
        bl, br, tl, tr = coords

        x_left = x_size - 0.5
        x_right = 0.5

        y_bottom = 0.5
        y_top = y_size - 0.5

        if orbit_direction == 'descending':
            x_left, x_right = x_right, x_left
            y_bottom, y_top = y_top, y_bottom

        gcps.extend([
            gdal.GCP(bl[1], bl[0], 0, x_left, y_bottom),
            gdal.GCP(br[1], br[0], 0, x_right, y_bottom),
            gdal.GCP(tl[1], tl[0], 0, x_left, y_top),
            gdal.GCP(tr[1], tr[0], 0, x_right, y_top),
        ])

    return gcps
